TEMPLATE_SOURCE_TYPE = "TemplateSourceType"

# stairlight.yaml
STAIRLIGHT_CONFIG_FILE_PREFIX = "stairlight"
STAIRLIGHT_CONFIG_INCLUDE_SECTION = "Include"
STAIRLIGHT_CONFIG_EXCLUDE_SECTION = "Exclude"
STAIRLIGHT_CONFIG_SETTING_SECTION = "Settings"

DEFAULT_TABLE_PREFIX = "DefaultTablePrefix"
FILE_SYSTEM_PATH = "FileSystemPath"
REGEX = "Regex"

# mapping.yaml
MAPPING_CONFIG_FILE_PREFIX = "mapping"
MAPPING_CONFIG_GLOBAL_SECTION = "Global"
MAPPING_CONFIG_MAPPING_SECTION = "Mapping"
MAPPING_CONFIG_METADATA_SECTION = "Metadata"

FILE_SUFFIX = "FileSuffix"
LABELS = "Labels"
MAPPING_PREFIX = "MappingPrefix"
PARAMETERS = "Parameters"
TABLE_NAME = "TableName"
TABLES = "Tables"

# GCS
BUCKET_NAME = "BucketName"
PROJECT_ID = "ProjectId"
URI = "Uri"

# Redash
DATABASE_URL_ENVIRONMENT_VARIABLE = "DatabaseUrlEnvironmentVariable"
DATA_SOURCE_NAME = "DataSourceName"
QUERY_IDS = "QueryIds"
QUERY_ID = "QueryId"



import enum
import re
from logging import getLogger
from typing import Iterator, Optional

from jinja2 import BaseLoader, Environment

from .. import config_key


class TemplateSourceType(enum.Enum):
    """SQL template source type"""

    FILE = "File"
    GCS = "GCS"
    REDASH = "Redash"

    def __str__(self):
        return self.name


class Template:
    """Base SQL template"""

    def __init__(
        self,
        mapping_config: dict,
        key: str,
        source_type: TemplateSourceType,
        bucket: Optional[str] = None,
        project: Optional[str] = None,
        default_table_prefix: Optional[str] = None,
    ):
        """SQL template

        Args:
            mapping_config (dict): Mapping configuration
            key (str): SQL file key
            source_type (SourceType): Source type
            bucket (Optional[str], optional):
                Bucket name where SQL file saved.Defaults to None.
            project (Optional[str], optional):
                Project name where SQL file saved.Defaults to None.
            default_table_prefix (Optional[str], optional):
                If project or dataset that configured table have are omitted,
                it will be complement this prefix. Defaults to None.
        """
        self._mapping_config = mapping_config
        self.key = key
        self.source_type = source_type
        self.bucket = bucket
        self.project = project
        self.default_table_prefix = default_table_prefix
        self.uri = ""

    def get_mapped_table_attributes_iter(self) -> Iterator[dict]:
        """Get mapped tables as iterator

        Yields:
            Iterator[dict]: Mapped table attributes
        """
        for mapping in self._mapping_config.get(
            config_key.MAPPING_CONFIG_MAPPING_SECTION
        ):
            has_suffix = False
            if self.key and mapping.get(config_key.FILE_SUFFIX):
                has_suffix = self.key.endswith(mapping.get(config_key.FILE_SUFFIX))
            if has_suffix or self.uri == mapping.get(config_key.URI):
                for table_attributes in mapping.get(config_key.TABLES):
                    yield table_attributes
                break

    def is_mapped(self) -> bool:
        """Check if the template is set to mapping configuration

        Returns:
            bool: Is set or not
        """
        result = False
        for _ in self.get_mapped_table_attributes_iter():
            result = True
            break
        return result

    @staticmethod
    def get_jinja_params(template_str) -> list:
        """Get jinja parameters

        Args:
            template_str (str): File string

        Returns:
            list: Jinja parameters
        """
        jinja_expressions = "".join(
            re.findall("{{[^}]*}}", template_str, re.IGNORECASE)
        )
        return re.findall("[^{} ]+", jinja_expressions, re.IGNORECASE)

    @staticmethod
    def render_by_base_loader(template_str: str, params: dict) -> str:
        jinja_template = Environment(loader=BaseLoader()).from_string(template_str)
        return jinja_template.render(params)

    def get_uri(self) -> str:
        """Get uri"""
        return ""

    def get_template_str(self) -> str:
        """Get template strings that read from template source"""
        return ""

    def render(self, params: dict) -> str:
        """Render SQL query string from a jinja template"""
        return ""


class TemplateSource:
    """SQL template source"""

    logger = getLogger(__name__)

    def __init__(self, stairlight_config: dict, mapping_config: dict) -> None:
        """SQL template source

        Args:
            stairlight_config (dict): Stairlight configuration
            mapping_config (dict): Mapping configuration
        """
        self._stairlight_config = stairlight_config
        self._mapping_config = mapping_config

    def search_templates_iter(self) -> Iterator[Template]:
        """Search SQL template files

        Yields:
            Iterator[SQLTemplate]: SQL template file attributes
        """
        pass

    def is_excluded(self, source_type: TemplateSourceType, key: str) -> bool:
        """Check if the specified file is out of scope

        Args:
            source_type (TemplateSourceType): SQL template source type
            key (str): SQL template file key

        Returns:
            bool: Return True if the specified file is out of scope
        """
        result = False
        exclude_list = self._stairlight_config.get(
            config_key.STAIRLIGHT_CONFIG_EXCLUDE_SECTION
        )
        if not exclude_list:
            return result
        for exclude in exclude_list:
            if source_type.value == exclude.get(
                config_key.TEMPLATE_SOURCE_TYPE
            ) and re.search(rf"{exclude.get(config_key.REGEX)}", key):
                result = True
                break
        return result



import os
import pathlib
import re
from typing import Iterator, Optional

from jinja2 import Environment, FileSystemLoader

from .. import config_key
from .base import Template, TemplateSource, TemplateSourceType


class FileTemplate(Template):
    def __init__(
        self,
        mapping_config: dict,
        key: str,
        source_type: Optional[TemplateSourceType] = TemplateSourceType.FILE,
        bucket: Optional[str] = None,
        project: Optional[str] = None,
        default_table_prefix: Optional[str] = None,
    ):
        super().__init__(
            mapping_config=mapping_config,
            key=key,
            source_type=source_type,
            bucket=bucket,
            project=project,
            default_table_prefix=default_table_prefix,
        )
        self.uri = self.get_uri()

    def get_uri(self) -> str:
        """Get uri from file path

        Returns:
            str: uri
        """
        return str(pathlib.Path(self.key).resolve())

    def get_template_str(self) -> str:
        """Get template string that read from a file in local file system

        Returns:
            str: Template string
        """
        with open(self.key) as f:
            return f.read()

    def render(self, params: dict) -> str:
        """Render SQL query string from a jinja template on local file system

        Args:
            params (dict): Jinja paramters

        Returns:
            str: SQL query string
        """
        if params:
            env = Environment(loader=FileSystemLoader(os.path.dirname(self.key)))
            jinja_template = env.get_template(os.path.basename(self.key))
            results = jinja_template.render(params)
        else:
            results = self.get_template_str()
        return results


class FileTemplateSource(TemplateSource):
    def __init__(
        self, stairlight_config: dict, mapping_config: dict, source_attributes: dict
    ) -> None:
        super().__init__(stairlight_config, mapping_config)
        self.source_attributes = source_attributes
        self.source_type = TemplateSourceType.FILE

    def search_templates_iter(self) -> Iterator[Template]:
        """Search SQL template files from local file system

        Args:
            source (dict): Source attributes of SQL template files

        Yields:
            Iterator[SQLTemplate]: SQL template file attributes
        """
        path_obj = pathlib.Path(self.source_attributes.get(config_key.FILE_SYSTEM_PATH))
        for p in path_obj.glob("**/*"):
            if (
                not re.fullmatch(
                    rf"{self.source_attributes.get(config_key.REGEX)}",
                    str(p),
                )
            ) or self.is_excluded(source_type=self.source_type, key=str(p)):
                self.logger.debug(f"{str(p)} is skipped.")
                continue
            yield FileTemplate(
                mapping_config=self._mapping_config,
                key=str(p),
                source_type=self.source_type,
                default_table_prefix=self.source_attributes.get(
                    config_key.DEFAULT_TABLE_PREFIX
                ),
            )



import os
from logging import getLogger
from typing import Iterator, Optional

from sqlalchemy import create_engine, text

from .. import config_key
from .base import Template, TemplateSource, TemplateSourceType

logger = getLogger(__name__)


class RedashTemplate(Template):
    def __init__(
        self,
        mapping_config: dict,
        query_id: int,
        query_name: str,
        query_str: str = None,
        data_source_name: str = None,
        source_type: Optional[TemplateSourceType] = TemplateSourceType.REDASH,
        key: Optional[str] = None,
        bucket: Optional[str] = None,
        project: Optional[str] = None,
        default_table_prefix: Optional[str] = None,
    ):
        super().__init__(
            mapping_config=mapping_config,
            key=query_id,
            source_type=source_type,
            bucket=bucket,
            project=project,
            default_table_prefix=default_table_prefix,
        )
        self.query_id = query_id
        self.query_str = query_str
        self.uri = query_name
        self.data_source_name = data_source_name

    def get_mapped_table_attributes_iter(self) -> Iterator[dict]:
        """Get mapped tables as iterator

        Yields:
            Iterator[dict]: Mapped table attributes
        """
        for mapping in self._mapping_config.get(
            config_key.MAPPING_CONFIG_MAPPING_SECTION
        ):
            if (
                mapping.get(config_key.QUERY_ID) == self.query_id
                and mapping.get(config_key.DATA_SOURCE_NAME) == self.data_source_name
            ):
                for table_attributes in mapping.get(config_key.TABLES):
                    yield table_attributes
                break

    def get_template_str(self) -> str:
        """Get template string that read from Redash
        Returns:
            str: Template string
        """
        return self.query_str

    def render(self, params: dict) -> str:
        """Render SQL query string from a jinja template on Redash queries
        Args:
            params (dict): Jinja paramters
        Returns:
            str: SQL query string
        """
        template_str = self.get_template_str()
        if params:
            results = self.render_by_base_loader(
                template_str=template_str, params=params
            )
        else:
            results = template_str
        return results


class RedashTemplateSource(TemplateSource):
    def __init__(
        self, stairlight_config: dict, mapping_config: dict, source_attributes: dict
    ) -> None:
        super().__init__(
            stairlight_config=stairlight_config, mapping_config=mapping_config
        )
        self.source_type = TemplateSourceType.REDASH
        self.source_attributes = source_attributes
        self.where_clause = []
        self.conditions = {
            "data_source": {
                "key": config_key.DATA_SOURCE_NAME,
                "query": "data_sources.name = :data_source",
                "parameters": self.source_attributes.get(config_key.DATA_SOURCE_NAME),
            },
            "query_ids": {
                "key": config_key.QUERY_IDS,
                "query": "queries.id IN :query_ids",
                "parameters": (
                    tuple(self.source_attributes.get(config_key.QUERY_IDS))
                    if self.source_attributes.get(config_key.QUERY_IDS)
                    else None
                ),
            },
        }

    def search_templates_iter(self) -> Iterator[Template]:
        results = self.get_queries_from_redash()
        for result in results:
            yield RedashTemplate(
                mapping_config=self._mapping_config,
                query_id=result[0],
                query_name=result[1],
                query_str=result[2],
                data_source_name=result[3],
            )

    def get_queries_from_redash(self) -> list:
        sql_file_name = "sql/redash_queries.sql"
        current_dir = os.path.dirname(os.path.abspath(__file__))
        query_text = text(
            self.build_query_string(path=f"{current_dir}/{sql_file_name}")
        )

        connection_str = self.get_connection_str()
        engine = create_engine(connection_str)
        queries = engine.execute(
            query_text,
            data_source=self.conditions.get("data_source").get("parameters"),
            query_ids=self.conditions.get("query_ids").get("parameters"),
        ).fetchall()

        return queries

    def build_query_string(self, path: str) -> str:
        base_query_string = self.read_query_from_file(path=path)
        for condition in self.conditions.values():
            if self.source_attributes.get(condition["key"]):
                self.where_clause.append(condition["query"])
        return base_query_string + "WHERE " + " AND ".join(self.where_clause)

    def read_query_from_file(self, path: str) -> str:
        with open(path, "r", encoding="utf-8") as f:
            return f.read()

    def get_connection_str(self) -> str:
        environment_variable_name = self.source_attributes.get(
            config_key.DATABASE_URL_ENVIRONMENT_VARIABLE
        )
        connection_str = os.environ.get(environment_variable_name)
        if not connection_str:
            logger.error(f"{environment_variable_name} is not found.")
        return connection_str



from typing import Iterator

from . import config_key, map_key
from .query import Query
from .source.base import Template, TemplateSource, TemplateSourceType
from .source.file import FileTemplateSource
from .source.gcs import GcsTemplateSource
from .source.redash import RedashTemplateSource


class Map:
    """Manages functions related to dependency map objects"""

    def __init__(
        self, stairlight_config: dict, mapping_config: dict, mapped: dict = None
    ) -> None:
        """Manages functions related to dependency map objects

        Args:
            stairlight_config (dict): Stairlight configuration
            mapping_config (dict): Mapping configuration
            mapped (dict, optional):
                Mapped file attributes when a mapping configuration file loaded.
                Defaults to {}.
        """
        if mapped:
            self.mapped = mapped
        else:
            self.mapped = {}
        self.unmapped: list[dict] = []
        self.stairlight_config = stairlight_config
        self.mapping_config = mapping_config

    def write(self) -> None:
        """Write a dependency map"""
        for template_source in self.find_template_source(
            stairlight_config=self.stairlight_config,
            mapping_config=self.mapping_config,
        ):
            self.write_by_template_source(template_source=template_source)

    @staticmethod
    def find_template_source(
        stairlight_config: dict, mapping_config: dict
    ) -> Iterator[TemplateSource]:
        """find template source

        Args:
            stairlight_config (dict): Stairlight configuration
            mapping_config (dict): Mapping configuration

        Yields:
            Iterator[TemplateSource]: Template source class
        """
        for source_attributes in stairlight_config.get(
            config_key.STAIRLIGHT_CONFIG_INCLUDE_SECTION
        ):
            template_source: TemplateSource = None
            template_source_type = source_attributes.get(
                config_key.TEMPLATE_SOURCE_TYPE
            )
            if template_source_type == TemplateSourceType.FILE.value:
                template_source = FileTemplateSource
            elif template_source_type == TemplateSourceType.GCS.value:
                template_source = GcsTemplateSource
            elif template_source_type == TemplateSourceType.REDASH.value:
                template_source = RedashTemplateSource
            else:
                print(f"Template source is not found: {type}")
                continue
            yield template_source(
                stairlight_config=stairlight_config,
                mapping_config=mapping_config,
                source_attributes=source_attributes,
            )

    def write_by_template_source(self, template_source: TemplateSource) -> None:
        """Write a dependency map"""
        for template in template_source.search_templates_iter():
            if not self.mapping_config:
                self.add_unmapped_params(template=template)
            elif template.is_mapped():
                for table_attributes in template.get_mapped_table_attributes_iter():
                    self.find_unmapped_params(
                        template=template, table_attributes=table_attributes
                    )
                    self.remap(template=template, table_attributes=table_attributes)
            else:
                self.add_unmapped_params(template=template)

    def remap(self, template: Template, table_attributes: dict) -> None:
        """Remap a dependency map

        Args:
            template (Template): SQL template
            table_attributes (dict): Table attributes from mapping configuration
        """
        query_str: str = template.render(
            params=self.get_combined_params(table_attributes)
        )
        query = Query(
            query_str=query_str,
            default_table_prefix=template.default_table_prefix,
        )

        downstairs: str = table_attributes.get(config_key.TABLE_NAME)
        mapping_labels: dict = table_attributes.get(config_key.LABELS)
        metadata: list[str] = self.mapping_config.get(
            config_key.MAPPING_CONFIG_METADATA_SECTION
        )

        if downstairs not in self.mapped:
            self.mapped[downstairs] = {}

        for upstairs_attributes in query.get_upstairs_attributes_iter():
            upstairs = upstairs_attributes[map_key.TABLE_NAME]

            if not self.mapped[downstairs].get(upstairs):
                self.mapped[downstairs][upstairs] = self.create_upstairs_value(
                    template=template,
                    mapping_labels=mapping_labels,
                    metadata=metadata,
                    upstairs=upstairs,
                )

            self.mapped[downstairs][upstairs][map_key.LINES].append(
                {
                    map_key.LINE_NUMBER: upstairs_attributes[map_key.LINE_NUMBER],
                    map_key.LINE_STRING: upstairs_attributes[map_key.LINE_STRING],
                }
            )

    def get_global_params(self) -> dict:
        """get global parameters in mapping.yaml

        Returns:
            dict: global parameters
        """
        global_params: dict = {}
        global_section: dict = self.mapping_config.get(
            config_key.MAPPING_CONFIG_GLOBAL_SECTION
        )
        if config_key.PARAMETERS in global_section:
            global_params = global_section.get(config_key.PARAMETERS)

        return global_params

    def get_combined_params(self, table_attributes: dict) -> dict:
        """return a combination of global parameters and table parameters

        Args:
            table_attributes (dict): table attributes

        Returns:
            dict: combined parameters
        """
        global_params: dict = self.get_global_params()
        table_params: dict = table_attributes.get(config_key.PARAMETERS, {})

        # Table parameters are prioritized over global parameters
        return {**global_params, **table_params}

    @staticmethod
    def create_upstairs_value(
        template: Template,
        mapping_labels: dict,
        metadata: "list[str]",
        upstairs: str,
    ) -> dict:
        """create upstairs table information

        Args:
            template (Template): Template class
            mapping_labels (dict): labels in mapping section
            metadata (list[str]): metadata
            upstairs (str): upstairs table name

        Returns:
            dict: upstairs table information
        """
        metadata_labels = []
        upstairs_values = {
            map_key.TEMPLATE_SOURCE_TYPE: template.source_type.value,
            map_key.KEY: template.key,
            map_key.URI: template.uri,
            map_key.LINES: [],
        }

        if template.source_type == TemplateSourceType.GCS:
            upstairs_values[map_key.BUCKET_NAME] = template.bucket
        elif template.source_type == TemplateSourceType.REDASH:
            upstairs_values[map_key.DATA_SOURCE_NAME] = template.data_source_name

        if metadata:
            metadata_labels = [
                m.get(config_key.LABELS)
                for m in metadata
                if m.get(config_key.TABLE_NAME) == upstairs
            ]
        if mapping_labels or metadata_labels:
            upstairs_values[map_key.LABELS] = {}

        if mapping_labels:
            upstairs_values[map_key.LABELS] = {
                **upstairs_values[map_key.LABELS],
                **mapping_labels,
            }

        if metadata_labels:
            upstairs_values[map_key.LABELS] = {
                **upstairs_values[map_key.LABELS],
                **metadata_labels[0],
            }
        return upstairs_values

    def add_unmapped_params(self, template: Template, params: list = None) -> None:
        """add to the list of unmapped params

        Args:
            template (Template): SQL template
            params (dict, optional): Jinja parameters
        """
        if not params:
            template_str = template.get_template_str()
            params = template.get_jinja_params(template_str)
        self.unmapped.append(
            {
                map_key.TEMPLATE: template,
                map_key.PARAMETERS: params,
            }
        )

    def find_unmapped_params(self, template: Template, table_attributes: dict) -> None:
        """find unmapped parameters in mapped files

        Args:
            template (Template): SQL template
            table_attributes (dict): Table attributes from mapping configuration
        """
        template_str: str = template.get_template_str()
        template_params: list = template.get_jinja_params(template_str)
        if not template_params:
            return

        mapped_params_dict: dict = self.get_combined_params(table_attributes)
        mapped_params: list = combine_nested_dict_keys(d=mapped_params_dict)
        diff_params: list = list(set(template_params) - set(mapped_params))

        if diff_params:
            self.add_unmapped_params(template=template, params=diff_params)


def combine_nested_dict_keys(d: dict, delimiter: str = ".") -> list:
    """combine nested dictionary keys and converts to a list

    Args:
        d (dict): dict

    Returns:
        list: results
    """
    results = []
    for key, value in d.items():
        if isinstance(value, dict):
            recursive_results = combine_nested_dict_keys(d=value)
            for recursive_result in recursive_results:
                concat = key + delimiter + recursive_result
                results.append(concat)
        else:
            results.append(key)
    return results



import enum
import json
import os
from logging import getLogger
from typing import Union

from . import config_key, map_key
from .config import Configurator
from .map import Map
from .source.gcs import GCS_URI_SCHEME, get_gcs_blob

logger = getLogger(__name__)


class ResponseType(enum.Enum):
    """Enum: Execution result type of up/down command"""

    TABLE: str = "table"
    FILE: str = "file"

    def __str__(self):
        return self.name


class SearchDirection(enum.Enum):
    """Enum: Search direction"""

    UP = "Upstairs"
    DOWN = "Downstairs"

    def __str__(self):
        return self.name


class Node:
    """A Node of singly-linked list"""

    def __init__(self, val, next=None):
        self.val = val
        self.next = None


class StairLight:
    """Table dependency detector"""

    def __init__(
        self, config_dir: str = ".", load_files: list = None, save_file: str = None
    ) -> None:
        """Table dependency detector

        Args:
            config_dir (str, optional):
                Stairlight configuration directory. Defaults to ".".
            load_files (list, optional):
                file names of loading results if load option set. Defaults to None.
            save_file (str, optional):
                A file name of saving results if save option set. Defaults to None.
        """
        self.load_files = load_files
        self.save_file = save_file
        self._configurator = Configurator(dir=config_dir)
        self._mapped = {}
        self._unmapped: list[dict] = []
        self.mapping_config = None
        self._stairlight_config = self._configurator.read(
            prefix=config_key.STAIRLIGHT_CONFIG_FILE_PREFIX
        )

    def create_map(self) -> None:
        if not self._stairlight_config:
            return

        if self.load_files:
            self.load_map()
        else:
            self._set_config()
            self._write_map()

        if self.save_file:
            self.save_map()

    @property
    def mapped(self) -> dict:
        """Return mapped

        Returns:
            dict: Mapped results
        """
        return self._mapped

    @property
    def unmapped(self) -> "list[dict]":
        """Return unmapped

        Returns:
            list[dict]: Unmapped results
        """
        return self._unmapped

    def has_stairlight_config(self) -> bool:
        """Exists stairlight configuration file or not

        Returns:
            bool: Exists stairlight configuration file or not
        """
        return self._stairlight_config is not None

    def _set_config(self) -> None:
        """Set config"""
        if not self._stairlight_config:
            logger.warning(
                f"{config_key.STAIRLIGHT_CONFIG_FILE_PREFIX}.y(a)ml' is not found."
            )
            return

        mapping_config_prefix = config_key.MAPPING_CONFIG_FILE_PREFIX
        if config_key.STAIRLIGHT_CONFIG_SETTING_SECTION in self._stairlight_config:
            settings = self._stairlight_config[
                config_key.STAIRLIGHT_CONFIG_SETTING_SECTION
            ]
            if config_key.MAPPING_PREFIX in settings:
                mapping_config_prefix = settings[config_key.MAPPING_PREFIX]
        self._mapping_config = self._configurator.read(prefix=mapping_config_prefix)

    def _write_map(self) -> None:
        """write a dependency map"""
        dependency_map = Map(
            stairlight_config=self._stairlight_config,
            mapping_config=self._mapping_config,
        )

        dependency_map.write()
        if self._mapping_config:
            self._mapped = dependency_map.mapped

        self._unmapped = dependency_map.unmapped

    def init(self, prefix: str = config_key.STAIRLIGHT_CONFIG_FILE_PREFIX) -> str:
        """Create Stairlight template file

        Args:
            prefix (str, optional):
                Template file prefix. Defaults to STAIRLIGHT_CONFIG_PREFIX.

        Returns:
            str: Template file name
        """
        return self._configurator.create_stairlight_template_file(prefix=prefix)

    def check(self, prefix: str = config_key.MAPPING_CONFIG_FILE_PREFIX) -> str:
        """Check mapped results and create a mapping template file

        Args:
            prefix (str, optional):
                Template file prefix. Defaults to MAPPING_CONFIG_PREFIX.

        Returns:
            str: Template file name
        """
        if self.load_files:
            logger.warning("Load option is used, skip checking.")
            return None
        elif not self._unmapped:
            return None

        return self._configurator.create_mapping_template_file(
            unmapped=self._unmapped, prefix=prefix
        )

    def save_map(self) -> None:
        """Save mapped results"""
        if self.save_file.startswith(GCS_URI_SCHEME):
            self.save_map_gcs()
        else:
            self.save_map_fs()

    def save_map_gcs(self) -> None:
        """Save mapped results to Google Cloud Storage"""
        blob = self.get_gcs_blob(self.save_file)
        blob.upload_from_string(
            data=json.dumps(obj=self._mapped, indent=2),
            content_type="application/json",
        )

    def save_map_fs(self) -> None:
        """Save mapped results to file system"""
        with open(self.save_file, "w") as f:
            json.dump(self._mapped, f, indent=2)

    def load_map(self) -> None:
        """Load mapped results"""
        for load_file in self.load_files:
            loaded_map = {}
            if load_file.startswith(GCS_URI_SCHEME):
                loaded_map = self.load_map_gcs(load_file=load_file)
            else:
                loaded_map = self.load_map_fs(load_file=load_file)

            if self._mapped:
                self._mapped = deep_merge(org=self._mapped, add=loaded_map)
            else:
                self._mapped = loaded_map

    def load_map_gcs(self, load_file: str) -> dict:
        """Load mapped results from Google Cloud Storage"""
        blob = get_gcs_blob(load_file)
        if not blob.exists():
            logger.error(f"{load_file} is not found.")
            exit()
        return json.loads(blob.download_as_string())

    def load_map_fs(self, load_file: str) -> dict:
        """Load mapped results from file system"""
        if not os.path.exists(load_file):
            logger.error(f"{load_file} is not found.")
            exit()
        with open(load_file) as f:
            return json.load(f)

    def up(
        self,
        table_name: str,
        recursive: bool = False,
        verbose: bool = False,
        response_type: str = ResponseType.TABLE.value,
    ) -> Union[list, dict]:
        """Search upstream nodes

        Args:
            table_name (str): Table name
            recursive (bool, optional): Search recursively or not. Defaults to False.
            verbose (bool, optional): Return verbose results or not. Defaults to False.
            response_type (str, optional):
                Response type. Defaults to ResponseType.TABLE.value.

        Returns:
            Union[list, dict]: [description]
        """
        return self.search(
            table_name=table_name,
            recursive=recursive,
            verbose=verbose,
            response_type=response_type,
            direction=SearchDirection.UP,
        )

    def down(
        self,
        table_name,
        recursive=False,
        verbose=False,
        response_type=ResponseType.TABLE.value,
    ) -> Union[list, dict]:
        """Search downstream nodes

        Args:
            table_name (str): Table name
            recursive (bool, optional): Search recursively or not. Defaults to False.
            verbose (bool, optional): Return verbose results or not. Defaults to False.
            response_type (str, optional):
                Response type. Defaults to ResponseType.TABLE.value.

        Returns:
            Union[list, dict]: [description]
        """
        return self.search(
            table_name=table_name,
            recursive=recursive,
            verbose=verbose,
            response_type=response_type,
            direction=SearchDirection.DOWN,
        )

    def search(
        self,
        table_name: str,
        recursive: bool,
        verbose: bool,
        response_type: str,
        direction: SearchDirection,
    ) -> Union[list, dict]:
        """Search nodes

        Args:
            table_name (str): Table name
            recursive (bool): Search recursively or not
            verbose (bool): Return verbose results or not
            response_type (str): Response type value
            direction (SearchDirection): Search direction

        Returns:
            Union[list, dict]: [description]
        """
        if verbose:
            return self.search_verbose(
                table_name=table_name,
                recursive=recursive,
                direction=direction,
                searched_tables=[],
                head=True,
            )

        if response_type in [type.value for type in ResponseType]:
            return self.search_plain(
                table_name=table_name,
                recursive=recursive,
                response_type=response_type,
                direction=direction,
                searched_tables=[],
                head=True,
            )

        return None

    def search_verbose(
        self,
        table_name: str,
        recursive: bool,
        direction: SearchDirection,
        searched_tables: list,
        head: bool,
    ) -> dict:
        """Search nodes and return verbose results

        Args:
            table_name (str): Table name
            recursive (bool): Search recursively or not
            direction (SearchDirection): Search direction
            searched_tables (list): a list of searched tables
            head (bool): Current position is head or not

        Returns:
            dict: Search results
        """
        relative_map = self.get_relative_map(table_name, direction)
        response = {table_name: {}}
        if not relative_map:
            return response

        if recursive:
            for next_table_name in relative_map.keys():
                if head:
                    searched_tables = [table_name]

                searched_tables.append(next_table_name)

                if is_cyclic(tables=searched_tables):
                    details = {
                        "table_name": table_name,
                        "next_table_name": next_table_name,
                        "searched_tables": searched_tables,
                    }
                    logger.warning(f"Circular reference detected!: {details}")
                    continue

                next_response = self.search_verbose(
                    table_name=next_table_name,
                    direction=direction,
                    recursive=recursive,
                    searched_tables=searched_tables,
                    head=False,
                )

                if not next_response.get(next_table_name):
                    continue

                relative_map[next_table_name] = {
                    **relative_map[next_table_name],
                    **next_response[next_table_name],
                }

        response[table_name][direction.value] = relative_map
        logger.debug(json.dumps(response, indent=2))
        return response

    def search_plain(
        self,
        table_name: str,
        recursive: bool,
        response_type: str,
        direction: SearchDirection,
        searched_tables: list,
        head: bool,
    ) -> list:
        """Search nodes and return simple results

        Args:
            table_name (str): Table name
            recursive (bool): Search recursively or not
            response_type (str): Response type value
            direction (SearchDirection): Search direction
            searched_tables (list): a list of searched tables
            head (bool): Current position is head or not

        Returns:
            list: Search results
        """
        relative_map = self.get_relative_map(table_name, direction)
        response = []
        if not relative_map:
            return response

        for next_table_name in relative_map.keys():
            if recursive:
                if head:
                    searched_tables = []
                    searched_tables.append(table_name)

                searched_tables.append(next_table_name)

                if is_cyclic(tables=searched_tables):
                    details = {
                        "table_name": table_name,
                        "next_table_name": next_table_name,
                        "searched_tables": searched_tables,
                    }
                    logger.info(f"Circular reference detected!: {details}")
                    continue

                next_response = self.search_plain(
                    table_name=next_table_name,
                    direction=direction,
                    recursive=recursive,
                    response_type=response_type,
                    searched_tables=searched_tables,
                    head=False,
                )
                response = response + next_response

            if response_type == ResponseType.TABLE.value:
                response.append(next_table_name)
            elif response_type == ResponseType.FILE.value:
                response.append(relative_map[next_table_name].get(map_key.URI))
            logger.debug(json.dumps(response, indent=2))

        return sorted(list(set(response)))

    def get_relative_map(self, table_name: str, direction: SearchDirection) -> dict:
        """Get a relative map for the specified direction

        Args:
            table_name (str): Table name
            direction (SearchDirection): Search direction

        Returns:
            dict: Relative map
        """
        relative_map = {}
        if direction == SearchDirection.UP:
            relative_map = self._mapped.get(table_name)
        elif direction == SearchDirection.DOWN:
            for key in [k for k, v in self._mapped.items() if v.get(table_name)]:
                relative_map[key] = self._mapped[key][table_name]
        return relative_map

    def get_tables_by_labels(self, targets: list) -> list:
        """Get tables to search by labels

        Args:
            targets (list): Target labels

        Returns:
            list: Tables to search
        """
        tables_to_search = []

        # "mapping" section in mapping.yaml
        for configurations in self._mapping_config.get(
            config_key.MAPPING_CONFIG_MAPPING_SECTION
        ):
            for table_attributes in configurations.get(config_key.TABLES):
                if self.is_target_found(
                    targets=targets,
                    labels=table_attributes.get(config_key.LABELS, {}),
                ):
                    tables_to_search.append(table_attributes[config_key.TABLE_NAME])

        # "metadata" section in mapping.yaml
        for table_attributes in self._mapping_config.get(
            config_key.MAPPING_CONFIG_METADATA_SECTION
        ):
            if self.is_target_found(
                targets=targets,
                labels=table_attributes.get(config_key.LABELS, {}),
            ):
                tables_to_search.append(table_attributes[config_key.TABLE_NAME])

        return tables_to_search

    @staticmethod
    def is_target_found(targets: list, labels: dict) -> bool:
        result = False
        found = 0
        for label_key, label_value in labels.items():
            for target in targets:
                target_key = target.split(":")[0]
                target_value = target.split(":")[1]
                if target_key == label_key and target_value == label_value:
                    found += 1
        if found == len(targets):
            result = True
        return result


def is_cyclic(tables: list) -> bool:
    """Floyd's cycle-finding algorithm

    Args:
        tables (list): Detected tables

    Returns:
        bool: Table dependencies are cyclic or not
    """
    nodes = {}
    for table in tables:
        if table not in nodes.keys():
            nodes[table] = Node(table)
    for i, table in enumerate(tables):
        if not nodes[table].next and i < len(tables) - 1:
            nodes[table].next = nodes[tables[i + 1]]
    slow = fast = nodes[tables[0]]
    while fast and fast.next:
        slow = slow.next
        fast = fast.next.next
        if slow == fast:
            return True
    return False


def deep_merge(org: dict, add: dict) -> dict:
    new: dict = org
    for add_key, add_value in add.items():
        org_value = org.get(add_key)
        if add_key not in org:
            new[add_key] = add_value
        elif isinstance(add_value, dict):
            new[add_key] = deep_merge(org=org_value, add=add_value)
    return new



import glob
import logging
import re
from collections import OrderedDict
from copy import deepcopy
from datetime import datetime, timezone

import yaml

from . import config_key, map_key
from .source.base import Template, TemplateSourceType

logger = logging.getLogger()


class Configurator:
    def __init__(self, dir: str) -> None:
        """Configuration class

        Args:
            path (str): Configuration file path
        """
        self.dir = dir

    def read(self, prefix: str) -> dict:
        """Read a configuration file

        Args:
            prefix (str): Configuration file name prefix

        Returns:
            dict: Results from reading configuration file
        """
        config = None
        pattern = f"^{self.dir}/{prefix}.ya?ml$"
        config_file = [
            p
            for p in glob.glob(f"{self.dir}/**", recursive=False)
            if re.fullmatch(pattern, p)
        ]
        if config_file:
            with open(config_file[0]) as file:
                config = yaml.safe_load(file)
        return config

    def create_stairlight_template_file(
        self, prefix: str = config_key.STAIRLIGHT_CONFIG_FILE_PREFIX
    ) -> str:
        """Create a Stairlight template file

        Args:
            prefix (str, optional): File prefix. Defaults to STAIRLIGHT_CONFIG_PREFIX.

        Returns:
            str: Created file name
        """
        template_file_name = f"{self.dir}/{prefix}.yaml"
        with open(template_file_name, "w") as f:
            yaml.add_representer(OrderedDict, self.represent_odict)
            yaml.dump(self.build_stairlight_config(), f)
        return template_file_name

    def create_mapping_template_file(
        self,
        unmapped: "list[dict]",
        prefix: str = config_key.MAPPING_CONFIG_FILE_PREFIX,
    ) -> str:
        """Create a mapping template file

        Args:
            unmapped (list[dict]): Unmapped results
            prefix (str, optional): File prefix. Defaults to MAPPING_CONFIG_PREFIX.

        Returns:
            str: Mapping template file
        """
        now = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
        template_file_name = f"{self.dir}/{prefix}_checked_{now}.yaml"

        with open(template_file_name, "w") as f:
            yaml.add_representer(
                data_type=OrderedDict, representer=self.represent_odict
            )
            yaml.dump(
                self.build_mapping_config(unmapped_templates=unmapped),
                stream=f,
            )
        return template_file_name

    @staticmethod
    def represent_odict(
        dumper: yaml.Dumper, odict: OrderedDict
    ) -> yaml.nodes.MappingNode:
        """Create a OrderedDict object for dumping a YAML file
        in order of OrderedDict"""
        return dumper.represent_mapping(
            tag="tag:yaml.org,2002:map", mapping=odict.items()
        )

    @staticmethod
    def build_stairlight_config() -> OrderedDict:
        """Create a OrderedDict object for file 'stairlight.config'

        Returns:
            OrderedDict: stairlight.config template
        """
        include_section_file = OrderedDict(
            {
                config_key.TEMPLATE_SOURCE_TYPE: TemplateSourceType.FILE.value,
                config_key.FILE_SYSTEM_PATH: None,
                config_key.REGEX: None,
                config_key.DEFAULT_TABLE_PREFIX: None,
            }
        )
        include_section_gcs = OrderedDict(
            {
                config_key.TEMPLATE_SOURCE_TYPE: TemplateSourceType.GCS.value,
                config_key.PROJECT_ID: None,
                config_key.BUCKET_NAME: None,
                config_key.REGEX: None,
                config_key.DEFAULT_TABLE_PREFIX: None,
            }
        )
        return OrderedDict(
            {
                config_key.STAIRLIGHT_CONFIG_INCLUDE_SECTION: [
                    include_section_file,
                    include_section_gcs,
                ],
                config_key.STAIRLIGHT_CONFIG_EXCLUDE_SECTION: [
                    OrderedDict(
                        {
                            config_key.TEMPLATE_SOURCE_TYPE: None,
                            config_key.DEFAULT_TABLE_PREFIX: None,
                        }
                    )
                ],
                config_key.STAIRLIGHT_CONFIG_SETTING_SECTION: {
                    config_key.MAPPING_PREFIX: config_key.MAPPING_CONFIG_FILE_PREFIX
                },
            }
        )

    def build_mapping_config(self, unmapped_templates: "list[dict]") -> OrderedDict:
        """Create a OrderedDict for mapping.config

        Args:
            unmapped (list[dict]): unmapped settings that Stairlight detects

        Returns:
            OrderedDict: mapping.config template
        """
        mapping_config_dict = OrderedDict(
            {
                config_key.MAPPING_CONFIG_GLOBAL_SECTION: [],
                config_key.MAPPING_CONFIG_MAPPING_SECTION: [],
            }
        )

        all_parameters: list[OrderedDict] = []
        global_parameters: dict = {}

        # Mapping section
        unmapped_template: dict
        for unmapped_template in unmapped_templates:
            template: Template = unmapped_template[map_key.TEMPLATE]
            values = OrderedDict(
                {
                    config_key.TEMPLATE_SOURCE_TYPE: template.source_type.value,
                }
            )

            if template.source_type == TemplateSourceType.FILE:
                values[config_key.FILE_SUFFIX] = template.key
            elif template.source_type == TemplateSourceType.GCS:
                values[config_key.URI] = template.uri
                values[config_key.BUCKET_NAME] = template.bucket
            elif template.source_type == TemplateSourceType.REDASH:
                values[config_key.QUERY_ID] = template.query_id
                values[config_key.DATA_SOURCE_NAME] = template.data_source_name

            # Tables
            values[config_key.TABLES] = [OrderedDict({config_key.TABLE_NAME: None})]
            if template.source_type == TemplateSourceType.REDASH:
                values[config_key.TABLES][0][config_key.TABLE_NAME] = template.uri

            # Parameters
            parameters: OrderedDict = None
            if map_key.PARAMETERS in unmapped_template:
                undefined_params: list[str] = unmapped_template.get(map_key.PARAMETERS)
                parameters = OrderedDict()
                for undefined_param in undefined_params:
                    splitted_params = undefined_param.split(".")
                    create_nested_dict(keys=splitted_params, results=parameters)

            if parameters:
                values[config_key.TABLES][0][config_key.PARAMETERS] = parameters
                if parameters in all_parameters:
                    global_parameters.update(parameters)
                else:
                    all_parameters.append(parameters)

            # Labels
            values[config_key.TABLES][0][config_key.LABELS] = OrderedDict(
                {"key": "value"}
            )

            mapping_config_dict[config_key.MAPPING_CONFIG_MAPPING_SECTION].append(
                values
            )

        # Global section
        mapping_config_dict[config_key.MAPPING_CONFIG_GLOBAL_SECTION] = OrderedDict(
            deepcopy({config_key.PARAMETERS: global_parameters})
        )

        # Metadata section
        mapping_config_dict[config_key.MAPPING_CONFIG_METADATA_SECTION] = [
            OrderedDict(
                {
                    config_key.TABLE_NAME: None,
                    config_key.LABELS: OrderedDict({"key": "value"}),
                }
            )
        ]

        return mapping_config_dict


def create_nested_dict(
    keys: list, results: OrderedDict, density: int = 0, default_value: any = None
) -> None:
    """create nested dict from list

    Args:
        keys (list): Dict keys
        results (OrderedDict): Nested dict
        density (int, optional): Density. Defaults to 0.
        default_value (any, optional): Default dict value. Defaults to None.
    """
    key = keys[density]
    if density < len(keys) - 1:
        if key not in results:
            results[key] = {}
        create_nested_dict(keys=keys, results=results[key], density=density + 1)
    else:
        results[key] = default_value



import re
from typing import Iterator, Optional

from google.cloud import storage

from .. import config_key
from .base import Template, TemplateSource, TemplateSourceType

GCS_URI_SCHEME = "gs://"


class GcsTemplate(Template):
    def __init__(
        self,
        mapping_config: dict,
        key: str,
        source_type: Optional[TemplateSourceType] = TemplateSourceType.GCS,
        bucket: Optional[str] = None,
        project: Optional[str] = None,
        default_table_prefix: Optional[str] = None,
    ):
        super().__init__(
            mapping_config=mapping_config,
            key=key,
            source_type=source_type,
            bucket=bucket,
            project=project,
            default_table_prefix=default_table_prefix,
        )
        self.uri = self.get_uri()

    def get_uri(self) -> str:
        """Get uri from file path

        Returns:
            str: uri
        """
        return f"{GCS_URI_SCHEME}{self.bucket}/{self.key}"

    def get_template_str(self) -> str:
        """Get template string that read from a file in GCS

        Returns:
            str: Template string
        """
        client = storage.Client(credentials=None, project=self.project)
        bucket = client.get_bucket(self.bucket)
        blob = bucket.blob(self.key)
        return blob.download_as_bytes().decode("utf-8")

    def render(self, params: dict) -> str:
        """Render SQL query string from a jinja template on GCS

        Args:
            params (dict): Jinja paramters

        Returns:
            str: SQL query string
        """
        template_str = self.get_template_str()
        if params:
            results = self.render_by_base_loader(
                template_str=template_str, params=params
            )
        else:
            results = template_str
        return results


class GcsTemplateSource(TemplateSource):
    def __init__(
        self, stairlight_config: dict, mapping_config: dict, source_attributes: dict
    ) -> None:
        super().__init__(
            stairlight_config=stairlight_config, mapping_config=mapping_config
        )
        self.source_attributes = source_attributes
        self.source_type = TemplateSourceType.GCS

    def search_templates_iter(self) -> Iterator[Template]:
        """Search SQL template files from GCS

        Args:
            source (dict): Source attributes of SQL template files

        Yields:
            Iterator[SQLTemplate]: SQL template file attributes
        """
        project = self.source_attributes.get(config_key.PROJECT_ID)
        client = storage.Client(credentials=None, project=project)
        bucket = self.source_attributes.get(config_key.BUCKET_NAME)
        blobs = client.list_blobs(bucket)
        for blob in blobs:
            if (
                not re.fullmatch(
                    rf"{self.source_attributes.get(config_key.REGEX)}",
                    blob.name,
                )
            ) or self.is_excluded(source_type=self.source_type, key=blob.name):
                self.logger.debug(f"{blob.name} is skipped.")
                continue
            yield GcsTemplate(
                mapping_config=self._mapping_config,
                key=blob.name,
                source_type=self.source_type,
                project=project,
                bucket=bucket,
                default_table_prefix=self.source_attributes.get(
                    config_key.DEFAULT_TABLE_PREFIX
                ),
            )


def get_gcs_blob(gcs_uri: str) -> storage.Blob:
    bucket_name = gcs_uri.replace(GCS_URI_SCHEME, "").split("/")[0]
    key = gcs_uri.replace(f"{GCS_URI_SCHEME}{bucket_name}/", "")

    client = storage.Client(credentials=None, project=None)
    bucket = client.get_bucket(bucket_name)
    return bucket.blob(key)



TABLE_NAME = "TableName"
TEMPLATE_SOURCE_TYPE = "TemplateSourceType"
KEY = "Key"
URI = "Uri"
LINES = "Lines"
LINE_NUMBER = "LineNumber"
LINE_STRING = "LineString"
BUCKET_NAME = "BucketName"
LABELS = "Labels"
DATA_SOURCE_NAME = "DataSourceName"

TEMPLATE = "Template"
PARAMETERS = "Parameters"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['stairlight', 'stairlight.source']

package_data = \
{'': ['*'], 'stairlight.source': ['sql/*']}

install_requires = \
['Jinja2>=2.10.3',
 'PyYAML>=5.0',
 'SQLAlchemy>=1.4.31,<2.0.0',
 'google-cloud-storage>=1.28.1,<2.0.0',
 'psycopg2>=2.9.3,<3.0.0']

entry_points = \
{'console_scripts': ['stairlight = stairlight.cli:main']}

setup_kwargs = {
    'name': 'stairlight',
    'version': '0.3.3',
    'description': 'Table-level data lineage tool',
    'long_description': '# Stairlight\n\n[![PyPi Version](https://img.shields.io/pypi/v/stairlight.svg?style=flat-square&logo=PyPi)](https://pypi.org/project/stairlight/)\n[![PyPi License](https://img.shields.io/pypi/l/stairlight.svg?style=flat-square)](https://pypi.org/project/stairlight/)\n[![PyPi Python Versions](https://img.shields.io/pypi/pyversions/stairlight.svg?style=flat-square)](https://pypi.org/project/stairlight/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square)](https://github.com/psf/black)\n[![CI](https://github.com/tosh2230/stairlight/actions/workflows/ci.yml/badge.svg)](https://github.com/tosh2230/stairlight/actions/workflows/ci.yml)\n\nStairlight is a table-level data lineage tool, detects table dependencies by SELECT queries.\n\nQueries can be read from following systems.\n\n- [Google Cloud Storage](https://cloud.google.com/storage)\n    - Mainly designed for use with [Google Cloud Composer](https://cloud.google.com/composer)\n- [Redash](https://redash.io/)\n- Local file system(with Python Pathlib module)\n\n## Installation\n\n```sh\n$ pip install stairlight\n```\n\n## Getting Started\n\nThere are 3 steps to use.\n\n```sh\n# Step 1: Initialize and set data location settings\n$ stairlight init\n\'./stairlight.yaml\' has created.\nPlease edit it to set your data sources.\n\n# Step 2: Map SQL queries and tables, and add metadata\n$ stairlight check\n\'./mapping_checked_yyyyMMddhhmmss.yaml\' has created.\nPlease map undefined tables and parameters, and append to your latest configuration file.\n\n# Step 3: Get a table dependency map\n$ stairlight\n```\n\n## Description\n\n### Input\n\n- SQL `SELECT` queries\n- Configuration files (YAML)\n    - stairlight.yaml: SQL query locations and include/exclude conditions.\n    - mapping.yaml: Mapping SQL queries and tables.\n\n### Output\n\n- Dependency map (JSON)\n\n    <details>\n\n    <summary>Example</summary>\n\n    ```json\n    {\n        "PROJECT_d.DATASET_e.TABLE_f": {\n            "PROJECT_j.DATASET_k.TABLE_l": {\n                "TemplateSourceType": "File",\n                "Key": "tests/sql/main/one_line_2.sql",\n                "Uri": "/foo/bar/stairlight/tests/sql/main/one_line_2.sql",\n                "Lines": [\n                    {\n                        "LineNumber": 1,\n                        "LineString": "SELECT * FROM PROJECT_j.DATASET_k.TABLE_l WHERE 1 = 1"\n                    }\n                ]\n            },\n            "PROJECT_C.DATASET_C.TABLE_C": {\n                "TemplateSourceType": "GCS",\n                "Key": "sql/cte/cte_multi_line.sql",\n                "Uri": "gs://stairlight/sql/cte/cte_multi_line.sql",\n                "Lines": [\n                    {\n                        "LineNumber": 6,\n                        "LineString": "        PROJECT_C.DATASET_C.TABLE_C"\n                    }\n                ],\n                "BucketName": "stairlight",\n                "Labels": {\n                    "Source": "gcs",\n                    "Test": "b"\n                }\n            }\n        },\n        "AggregateSales": {\n            "PROJECT_e.DATASET_e.TABLE_e": {\n                "TemplateSourceType": "Redash",\n                "Key": 5,\n                "Uri": "AggregateSales",\n                "Lines": [\n                    {\n                        "LineNumber": 1,\n                        "LineString": "SELECT service, SUM(total_amount) FROM PROJECT_e.DATASET_e.TABLE_e GROUP BY service"\n                    }\n                ],\n                "DataSourceName": "BigQuery",\n                "Labels": {\n                    "Category": "Sales"\n                }\n            }\n        },\n    }\n    ```\n\n    </details>\n\n## Configuration\n\nConfiguration files can be found [here](https://github.com/tosh2230/stairlight/tree/main/tests/config), used for unit test in CI.\n\n### stairlight.yaml\n\n\'stairlight.yaml\' is for setting up Stairlight itself.\n\nIt is responsible for specifying the destination of SQL queries to be read, and for specifying data sources.\n\n```yaml\nInclude:\n  - TemplateSourceType: File\n    FileSystemPath: "./tests/sql"\n    Regex: ".*/*.sql$"\n    DefaultTablePrefix: "PROJECT_A"\n  - TemplateSourceType: GCS\n    ProjectId: null\n    BucketName: stairlight\n    Regex: "^sql/.*/*.sql$"\n    DefaultTablePrefix: "PROJECT_A"\n  - TemplateSourceType: Redash\n    DatabaseUrlEnvironmentVariable: REDASH_DATABASE_URL\n    DataSourceName: BigQuery\n    QueryIds:\n      - 1\n      - 3\n      - 5\nExclude:\n  - TemplateSourceType: File\n    Regex: "main/exclude.sql$"\nSettings:\n  MappingPrefix: "mapping"\n```\n\n### mapping.yaml\n\n\'mapping.yaml\' is used to define relationships between input queries and tables.\n\nA template of this file can be created by `check` command, based on the configuration of \'stairlight.yaml\'.\n\n```yaml\nGlobal:\n  Parameters:\n    DESTINATION_PROJECT: stairlight\n    params:\n      PROJECT: 1234567890\n      DATASET: public\n      TABLE: taxirides\nMapping:\n  - TemplateSourceType: File\n    FileSuffix: "tests/sql/main/union_same_table.sql"\n    Tables:\n      - TableName: "test_project.beam_streaming.taxirides_aggregation"\n  - TemplateSourceType: GCS\n    Uri: "gs://stairlight/sql/one_line/one_line.sql"\n    Tables:\n      - TableName: "PROJECT_a.DATASET_b.TABLE_c"\n  - TemplateSourceType: Redash\n    QueryId: 5\n    DataSourceName: metadata\n    Tables:\n      - TableName: Copy of (#4) New Query\n        Parameters:\n          table: dashboards\n        Labels:\n          Category: Redash test\nMetadata:\n  - TableName: "PROJECT_A.DATASET_A.TABLE_A"\n    Labels:\n      Source: Null\n      Test: a\n```\n\n#### Global Section\n\nThis section is for global configurations.\n\n`Parameters` attribute is used to set common parameters. If conflicts has occurred with `Parameters` attributes in mapping section, mapping section\'s parameters will be used in preference to global.\n\n#### Mapping Section\n\nMapping section is used to define relationships between queries and tables that created as a result of query execution.\n\n`Parameters` attribute allows you to reflect settings in [jinja](https://jinja.palletsprojects.com/) template variables embedded in queries. If multiple settings are applied to a query using jinja template, the query will be read as if there were the same number of queries as the number of settings.\n\n#### Metadata Section\n\nThis section is mainly used to set metadata to tables appears only in queries.\n\n## Command and Option\n\n```txt\n$ stairlight --help\nusage: stairlight [-h] [-c CONFIG] [--save SAVE] [--load LOAD] {init,check,up,down} ...\n\nA table-level data lineage tool, detects table dependencies by SELECT queries.\nWithout positional arguments, return a table dependency map as JSON format.\n\npositional arguments:\n  {init,check,up,down}\n    init                create new Stairlight configuration file\n    check               create new configuration file about undefined mappings\n    up                  return upstairs ( table | SQL file ) list\n    down                return downstairs ( table | SQL file ) list\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set Stairlight configuration directory\n  -q, --quiet           keep silence\n  --save SAVE           file path where results will be saved(File system or GCS)\n  --load LOAD           file path in which results are saved(File system or GCS), can be specified multiple times\n```\n\n### init\n\n`init` creates a new Stairlight configuration file.\n\n```txt\n$ stairlight init --help\nusage: stairlight init [-h] [-c CONFIG]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set Stairlight configuration directory.\n  -q, --quiet           keep silence\n```\n\n### check\n\n`check` creates new configuration file about undefined mappings.\nThe option specification is the same as `init`.\n\n### up\n\n`up` outputs a list of tables or SQL files located upstream from the specified table.\n\n- Use table(`-t`, `--table`) or label(`-l`, `--label`) option to specify tables to search.\n- Recursive option(`-r`, `--recursive`) is set, Stairlight will find tables recursively and output as a list.\n- Verbose option(`-v`, `--verbose`) is set, Stairlight will add detailed information and output it as a dict.\n\n```txt\n$ stairlight up --help\nusage: stairlight up [-h] [-c CONFIG] [--save SAVE] [--load LOAD] (-t TABLE | -l LABEL) [-o {table,file}]\n                     [-v] [-r]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set Stairlight configuration directory\n  -q, --quiet           keep silence\n  --save SAVE           file path where results will be saved(File system or GCS)\n  --load LOAD           file path in which results are saved(File system or GCS), can be specified multiple times\n  -t TABLE, --table TABLE\n                        table names that Stairlight searches for, can be specified\n                        multiple times. e.g. -t PROJECT_a.DATASET_b.TABLE_c -t\n                        PROJECT_d.DATASET_e.TABLE_f\n  -l LABEL, --label LABEL\n                        labels set for the table in mapping configuration, can be\n                        specified multiple times. The separator between key and value\n                        should be a colon(:). e.g. -l key_1:value_1 -l key_2:value_2\n  -o {table,file}, --output {table,file}\n                        output type\n  -v, --verbose         return verbose results\n  -r, --recursive       search recursively\n```\n\n### down\n\n`down` outputs a list of tables or SQL files located downstream from the specified table.\nThe option specification is the same as `up`.\n\n## Use as a library\n\nStairlight can also be used as a library.\n\n[tosh2230/stairlight-app](https://github.com/tosh2230/stairlight-app) is a sample web application rendering table dependency graph with Stairlight, using Graphviz, Streamlit and Google Cloud Run.\n',
    'author': 'tosh2230',
    'author_email': 'rev.to12@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
